@@ -28,6 +28,7 @@ from ._util import (
     NOTICE_LEVEL,
     install_coloredlogs,
     parse_byte_size,
+    path_really_within,
 )
 from ._util import StructuredLogMessage as _
 
@@ -447,6 +448,8 @@ def runner(
         level = VERBOSE_LEVEL
     if kwargs["debug"]:
         level = logging.DEBUG
+    else:
+        logging.raiseExceptions = False
     if kwargs["no_color"]:
         # picked up by _util.install_coloredlogs()
         os.environ["NO_COLOR"] = os.environ.get("NO_COLOR", "")
@@ -462,14 +465,14 @@ def runner(
     cfg = runtime.config.Loader(logger, filenames=cfg_arg)
     cfg_overrides = {
         "scheduler": {},
-        "task_io": {},
+        "file_io": {},
         "task_runtime": {},
         "download_cache": {},
     }
     if max_tasks is not None:
         cfg_overrides["call_concurrency"] = max_tasks
     if copy_input_files:
-        cfg_overrides["task_io"]["copy_input_files"] = copy_input_files
+        cfg_overrides["file_io"]["copy_input_files"] = copy_input_files
     if as_me:
         cfg_overrides["task_runtime"]["as_user"] = as_me
     if runtime_defaults:
@@ -492,6 +495,28 @@ def runner(
     cfg.override(cfg_overrides)
     cfg.log_all()
 
+    # check root
+    if not path_really_within((run_dir or os.getcwd()), cfg["file_io"]["root"]):
+        logger.error(
+            _(
+                "working directory or --dir must be within the configured `file_io.root' directory",
+                dir=(run_dir or os.getcwd()),
+                root=cfg["file_io"]["root"],
+            )
+        )
+        sys.exit(2)
+    if (
+        cfg["download_cache"].get_bool("get") or cfg["download_cache"].get_bool("put")
+    ) and not path_really_within(cfg["download_cache"]["dir"], cfg["file_io"]["root"]):
+        logger.error(
+            _(
+                "configuration error: 'download_cache.dir' must be within the `file_io.root' directory",
+                dir=cfg["download_cache"]["dir"],
+                root=cfg["file_io"]["root"],
+            )
+        )
+        sys.exit(2)
+
     # load WDL document
     doc = load(uri, path or [], check_quant=check_quant, read_source=read_source)
 
@@ -499,8 +524,16 @@ def runner(
     def file_found(fn):
         return runtime.download.able(cfg, fn) or os.path.isfile(fn)
 
+    eff_root = cfg["file_io"]["root"] if not cfg["file_io"].get_bool("copy_input_files") else "/"
+
     target, input_env, input_json = runner_input(
-        doc, inputs, input_file, empty, task=task, file_found=file_found
+        doc,
+        inputs,
+        input_file,
+        empty,
+        task=task,
+        file_found=file_found,
+        root=eff_root,  # if copy_input_files is set, then input files need not reside under the configured root
     )
 
     if json_only:
@@ -542,7 +575,7 @@ def runner(
     rerun_sh = f"pushd {shellquote(os.getcwd())} && miniwdl {' '.join(shellquote(t) for t in sys.argv[1:])}; popd"
 
     # initialize local Docker Swarm
-    runtime.task.LocalSwarmContainer.global_init(cfg, logger)
+    runtime.task.SwarmContainer.global_init(cfg, logger)
 
     # run & log any errors
     rundir = None
@@ -552,17 +585,11 @@ def runner(
         if error_json:
             print(json.dumps(runtime.error_json(exn), indent=2))
         exit_status = 2
-        pos = None
-        msg = None
         from_rundir = None
         while isinstance(exn, runtime.RunFailed):
             exn_rundir = getattr(exn, "run_dir")
             rundir = rundir or exn_rundir
             from_rundir = exn_rundir
-            exn_pos = getattr(exn, "pos", None)
-            if isinstance(exn_pos, SourcePosition):
-                pos = exn_pos
-            msg = str(exn)
             exn = exn.__cause__
         if isinstance(exn, runtime.CommandFailed):
             exit_status = (lambda v: v if v else exit_status)(getattr(exn, "exit_status", 0))
@@ -575,10 +602,10 @@ def runner(
             info["dir"] = rundir
         if from_rundir and from_rundir != rundir:
             info["from_dir"] = from_rundir
-        if not msg:
-            msg = str(exn)
-            if "message" in info:
-                del info["message"]  # redundant
+        msg = str(exn)
+        if "message" in info:
+            msg = info["message"]
+            del info["message"]
         logger.error(_(msg, **info))
         if kwargs["debug"]:
             raise
@@ -641,7 +668,9 @@ def runner_input_completer(prefix, parsed_args, **kwargs):
         return available_input_names
 
 
-def runner_input(doc, inputs, input_file, empty, task=None, check_required=True, file_found=None):
+def runner_input(
+    doc, inputs, input_file, empty, task=None, check_required=True, file_found=None, root="/"
+):
     """
     - Determine the target workflow/task
     - Check types of supplied inputs
@@ -668,7 +697,7 @@ def runner_input(doc, inputs, input_file, empty, task=None, check_required=True,
     # build up an values env of the provided inputs
     available_inputs = target.available_inputs
     input_env = runner_input_json_file(
-        available_inputs, (target.name if isinstance(target, Workflow) else ""), input_file
+        available_inputs, (target.name if isinstance(target, Workflow) else ""), input_file, root
     )
 
     # set explicitly empty arrays
@@ -708,7 +737,7 @@ def runner_input(doc, inputs, input_file, empty, task=None, check_required=True,
             )
 
         # create a Value based on the expected type
-        v = runner_input_value(s_value, decl.type, file_found)
+        v = runner_input_value(s_value, decl.type, file_found, root)
 
         # insert value into input_env
         try:
@@ -742,7 +771,7 @@ def runner_input(doc, inputs, input_file, empty, task=None, check_required=True,
     )
 
 
-def runner_input_json_file(available_inputs, namespace, input_file):
+def runner_input_json_file(available_inputs, namespace, input_file, root):
     """
     Load user-supplied inputs JSON file, if any
     """
@@ -767,8 +796,14 @@ def runner_input_json_file(available_inputs, namespace, input_file):
 
         def absolutify_files(v: Value.Base) -> Value.Base:
             if isinstance(v, Value.File):
-                if "://" not in v.value and not os.path.isabs(v.value):
-                    v.value = os.path.normpath(os.path.join(os.getcwd(), v.value))
+                if "://" not in v.value:
+                    if not os.path.isabs(v.value):
+                        v.value = os.path.normpath(os.path.join(os.getcwd(), v.value))
+                    if not path_really_within(v.value, root):
+                        die(
+                            f"all input files must be located within the configured `file_io.root' directory `{root}' unlike `{v.value}'"
+                        )
+                        sys.exit(2)
             for ch in v.children:
                 absolutify_files(ch)
             return v
@@ -819,7 +854,7 @@ def is_constant_expr(expr):
     return False
 
 
-def runner_input_value(s_value, ty, file_found):
+def runner_input_value(s_value, ty, file_found, root):
     """
     Given an input value from the command line (right-hand side of =) and the
     WDL type of the corresponding input decl, create an appropriate Value.
@@ -830,6 +865,10 @@ def runner_input_value(s_value, ty, file_found):
         fn = os.path.expanduser(s_value)
         if os.path.isfile(fn):
             fn = os.path.abspath(fn)
+            if not path_really_within(fn, root):
+                die(
+                    f"all input files must be located within the configured `file_io.root' directory `{root}' unlike `{fn}'"
+                )
         elif not (file_found and file_found(fn)):  # maybe URI
             die("File not found: " + fn)
         return Value.File(fn)
@@ -847,7 +886,9 @@ def runner_input_value(s_value, ty, file_found):
         ty.item_type, (Type.String, Type.File, Type.Int, Type.Float)
     ):
         # just produce a length-1 array, to be combined ex post facto
-        return Value.Array(ty.item_type, [runner_input_value(s_value, ty.item_type, file_found)])
+        return Value.Array(
+            ty.item_type, [runner_input_value(s_value, ty.item_type, file_found, root)]
+        )
     return die(
         "No command-line support yet for inputs of type {}; workaround: specify in JSON file with --input".format(
             str(ty)
@@ -944,21 +985,20 @@ def run_self_test(**kwargs):
         assert len(outputs["hello_caller.messages"]) == 2
         assert outputs["hello_caller.messages"][0].rstrip() == "Hello, Alyssa P. Hacker!"
         assert outputs["hello_caller.messages"][1].rstrip() == "Hello, Ben Bitdiddle!"
-    except SystemExit as exn:
-        assert getattr(exn, "code") == 0  # because of --debug
-    except:
-        atexit.register(
-            lambda: print(
-                "* Hint: ensure Docker is installed & running"
-                + (
-                    ", and user has permission to control it per https://docs.docker.com/install/linux/linux-postinstall/#manage-docker-as-a-non-root-user"
-                    if platform.system() != "Darwin"
-                    else "; and on macOS override the environment variable TMPDIR=/tmp/"
-                ),
-                file=sys.stderr,
+    except BaseException as exn:
+        if not (isinstance(exn, SystemExit) and getattr(exn, "code") == 0):
+            atexit.register(
+                lambda: print(
+                    "* Hint: ensure Docker is installed & running"
+                    + (
+                        ", and user has permission to control it per https://docs.docker.com/install/linux/linux-postinstall/#manage-docker-as-a-non-root-user"
+                        if platform.system() != "Darwin"
+                        else "; and on macOS override the environment variable TMPDIR=/tmp/"
+                    ),
+                    file=sys.stderr,
+                )
             )
-        )
-        raise
+            raise exn
 
     print("miniwdl run_self_test OK", file=sys.stderr)
     if os.geteuid() == 0:
@@ -1215,138 +1255,178 @@ def ensure_cromwell_jar(jarfile=None):
 def fill_localize_subparser(subparsers):
     localize_parser = subparsers.add_parser(
         "localize",
-        help="Download URIs found in Cromwell-style input JSON and rewrite",
-        description=f"Download URIs found in Cromwell-style input JSON, and rewrite it with the local filenames.",
+        help="Download URI input files to local cache",
+        description=f"Prime the local file download cache with URI File inputs found in Cromwell-style input JSON",
     )
     localize_parser.add_argument(
-        "wdlfile", metavar="DOC.wdl", type=str, help="WDL document filename/URI"
+        "wdlfile",
+        metavar="DOC.wdl",
+        type=str,
+        help="WDL document filename/URI",
+        default=None,
+        nargs="?",
     )
     localize_parser.add_argument(
         "infile",
         metavar="INPUT.json",
         type=str,
         help="input JSON filename (- for standard input) or literal object",
-    )
-    localize_parser.add_argument(
-        "name",
-        metavar="NAME",
-        type=str,
-        nargs="?",
         default=None,
-        help="short name to include in local paths (default: basename of JSON file)",
-    )
-    localize_parser.add_argument(
-        "-d",
-        "--dir",
-        metavar="DIR",
-        dest="run_dir",
-        help="base directory in which to store downloaded files",
-    )
-    localize_parser.add_argument(
-        "-o",
-        metavar="LOCAL.json",
-        type=str,
-        dest="outfile",
-        help="write transformed JSON to file instead of standard output",
+        nargs="?",
     )
     localize_parser.add_argument(
         "--task",
         metavar="TASK_NAME",
         help="name of task (for WDL documents with multiple tasks & no workflow)",
     )
+    localize_parser.add_argument(
+        "--uri",
+        metavar="URI",
+        action="append",
+        help="additional URI to process; or, omit WDL & JSON and just specify one or more --uri",
+    )
+    localize_parser.add_argument(
+        "--no-cache",
+        action="store_true",
+        help="if a URI is already cached, re-download and replace it",
+    )
+    localize_parser.add_argument(
+        "--cfg",
+        metavar="FILE",
+        type=str,
+        default=None,
+        help="configuration file to load (in preference to file named by MINIWDL_CFG environment, or XDG_CONFIG_{HOME,DIRS}/miniwdl.cfg)",
+    )
     return localize_parser
 
 
 def localize(
-    wdlfile, infile, name=None, outfile=None, task=None, path=None, check_quant=True, **kwargs
+    wdlfile=None,
+    infile=None,
+    uri=None,
+    no_cache=False,
+    task=None,
+    cfg=None,
+    path=None,
+    check_quant=True,
+    **kwargs,
 ):
-    # load WDL document
-    doc = load(wdlfile, path or [], check_quant=check_quant, read_source=read_source)
-
-    # parse the provided input JSON
     logging.basicConfig(level=NOTICE_LEVEL)
     logger = logging.getLogger("miniwdl-localize")
-    logger.critical(
-        "DEPRECATION NOTICE: `miniwdl localize` will soon be retired, superseded by the configurable URI download cache in >= v0.7.x"
-    )
-    cfg = runtime.config.Loader(logger)
+    install_coloredlogs(logger)
 
-    def file_found(fn):
-        return runtime.download.able(cfg, fn) or os.path.isfile(fn)
-
-    target, input_env, input_json = runner_input(
-        doc, [], infile, [], task=task, check_required=False, file_found=file_found
+    cfg_arg = None
+    if cfg:
+        assert os.path.isfile(cfg), "--cfg file not found"
+        cfg_arg = [cfg]
+    cfg = runtime.config.Loader(logger, filenames=cfg_arg)
+    cache_cfg = cfg["download_cache"]
+    original_get = cache_cfg.get_bool("get")
+    if original_get and no_cache:
+        cfg.override({"download_cache": {"get": False}})
+    logger.notice(
+        _(
+            "effective configuration",
+            put=cache_cfg.get_bool("put"),
+            get=cache_cfg.get_bool("get"),
+            dir=cache_cfg["dir"],
+            ignore_query=cache_cfg.get_bool("ignore_query"),
+            enable_patterns=cache_cfg.get_list("enable_patterns"),
+            disable_patterns=cache_cfg.get_list("disable_patterns"),
+        )
     )
 
-    # read input JSON
-    name = name or os.path.basename(infile).split(".")[0]
+    uri = uri or []
+    uri = set(uri)
 
-    # scan for Files that appear to be downloadable URIs
-    def scan(x):
-        if isinstance(x, Value.File) and runtime.download.able(cfg, x.value):
-            yield x.value
-        for y in x.children:
-            yield from scan(y)
+    if infile:
+        # load WDL document
+        doc = load(wdlfile, path or [], check_quant=check_quant, read_source=read_source)
 
-    uris = set()
-    for b in input_env:
-        uris |= set(scan(b.value))
+        def file_found(fn):
+            return runtime.download.able(cfg, fn) or os.path.isfile(fn)
 
-    if uris:
-        # initialize local Docker Swarm
-        runtime.task.LocalSwarmContainer.global_init(cfg, logger)
-
-        # cheesy trick: provide the list of URIs as File inputs to a dummy workflow, causing the
-        # runtime to download them
-        localizer_wdl = (
-            """
-            version 1.0
-            workflow localize_%s {
-                input {
-                    Array[File] uris
-                }
-                output {
-                    Array[File] files = uris
-                }
-            }
-            """
-            % name
-        )
-        localizer = parse_document(localizer_wdl)
-        localizer.typecheck()
-        cfg = runtime.config.Loader(logger)
-        subdir, outputs = runtime.run(
-            cfg,
-            localizer.workflow,
-            values_from_json({"uris": list(uris)}, localizer.workflow.available_inputs),
+        target, input_env, input_json = runner_input(
+            doc, [], infile, [], task=task, check_required=False, file_found=file_found
         )
 
-        # recover the mapping of URIs to downloaded files
-        uri_to_file = {}
-        assert isinstance(outputs["files"], Value.Array)
-        for uri, elt in zip(uris, outputs["files"].value):
-            assert isinstance(elt, Value.File) and os.path.isfile(elt.value)
-            uri_to_file[uri] = elt.value
-
-        # rewrite the input Env to replace URIs with filenames
-        def rewrite(x):
-            if isinstance(x, Value.File) and x.value in uri_to_file:
-                x.value = uri_to_file[x.value]
+        # scan for Files that appear to be downloadable URIs
+        def scan(x):
+            if isinstance(x, Value.File) and runtime.download.able(cfg, x.value):
+                yield x.value
             for y in x.children:
-                rewrite(y)
+                yield from scan(y)
 
         for b in input_env:
-            rewrite(b.value)
+            uri |= set(scan(b.value))
 
-    # write out the possibly-modified JSON
-    result_json = values_to_json(
-        input_env, namespace=(target.name if isinstance(target, Workflow) else "")
+    if not uri:
+        logger.warning(
+            "nothing to do; if inputs use special URI schemes, make sure necessary downloader plugin(s) are installed and enabled"
+        )
+        sys.exit(0)
+
+    if not cache_cfg.get_bool("put"):
+        logger.error(
+            'configuration section "download_cache", option "put" (env MINIWDL__DOWNLOAD_CACHE__PUT) must be true for this operation to be effective'
+        )
+        sys.exit(2)
+
+    if not path_really_within(cfg["download_cache"]["dir"], cfg["file_io"]["root"]):
+        logger.error(
+            _(
+                "configuration error: `download_cache.dir' must be within the `file_io.root' directory",
+                dir=cfg["download_cache"]["dir"],
+                root=cfg["file_io"]["root"],
+            )
+        )
+        sys.exit(2)
+
+    cache = runtime.cache.CallCache(cfg, logger)
+    disabled = [u for u in uri if not cache.download_path(u)]
+    if disabled:
+        logger.notice(_("URIs found but not cacheable per configuration", uri=disabled))
+    uri = list(uri - set(disabled))
+
+    if not uri:
+        logger.warning("nothing to do; check configured enable_patterns and disable_patterns")
+        sys.exit(0)
+    logger.notice(_("starting downloads", uri=uri))
+
+    # initialize local Docker Swarm
+    runtime.task.SwarmContainer.global_init(cfg, logger)
+
+    # cheesy trick: provide the list of URIs as File inputs to a dummy workflow, causing the
+    # runtime to download & cache them
+    localizer_wdl = """
+        version 1.0
+        workflow localize {
+            input {
+                Array[File] uris
+            }
+            output {
+                Array[File] files = uris
+            }
+        }
+        """
+    localizer = parse_document(localizer_wdl)
+    localizer.typecheck()
+    cfg = runtime.config.Loader(logger)
+    subdir, outputs = runtime.run(
+        cfg,
+        localizer.workflow,
+        values_from_json({"uris": uri}, localizer.workflow.available_inputs),
+        run_dir=os.environ.get("TMPDIR", "/tmp"),
+        _cache=cache,
     )
-    if outfile in [None, "", "-"]:
-        print(json.dumps(result_json, indent=2))
-    else:
-        with open(outfile, "w") as outp:
-            print(json.dumps(result_json, indent=2), file=outp)
+
+    logger.notice(
+        _("success", files=[os.path.realpath(p) for p in values_to_json(outputs)["files"]])
+    )
+    if not original_get:
+        logger.warning(
+            """future runs won't use the cache unless configuration section "download_cache", key "get" (env MINIWDL__DOWNLOAD_CACHE__GET) is set to true"""
+        )
 
 
 def die(msg, status=2):
