@@ -5,50 +5,35 @@ referenced therein, and updates their access timestamps (atime).
 """
 
 import os
-import time
-import fcntl
 import logging
 import threading
 from typing import Iterator, Dict, Any, Optional, Set, List, IO
-from contextlib import contextmanager, ExitStack
 from urllib.parse import urlparse, urlunparse
 from fnmatch import fnmatchcase
 from . import config
 from .. import Env, Value, Type
-from .._util import StructuredLogMessage as _
+from .._util import StructuredLogMessage as _, FlockHolder
 
 
 class CallCache:
     _cfg: config.Loader
-    _lock: threading.Lock
-    _flocked_files: Set[str]
-    _flocks: List[ExitStack]
+    _flocker: FlockHolder
+    _logger: logging.Logger
 
-    def __init__(self, cfg: config.Loader):
+    def __init__(self, cfg: config.Loader, logger: logging.Logger):
         self._cfg = cfg
-        self._lock = threading.Lock()
-        self._flocked_files = set()
-        self._flocks = []
-
-    def _flock(self, filenames: List[str]) -> None:
-        # open shared flocks on the specified filenames (all or none)
-        filenames2 = set(os.path.realpath(fn) for fn in filenames)
-        with self._lock:
-            filenames2 = filenames2 - self._flocked_files
-            if filenames2:
-                with ExitStack() as stack:
-                    for fn in filenames2:
-                        stack.enter_context(_open_and_flock(fn))  # pylint: disable=no-member
-                    self._flocked_files |= filenames2
-                    self._flocks.append(stack.pop_all())  # pylint: disable=no-member
+        self._logger = logger.getChild("CallCache")
+        self._flocker = FlockHolder(self._logger)
+        self._flocker.__enter__()
 
     def __del__(self):
-        with self._lock:
-            for lock in self._flocks:
-                lock.close()
+        self._flocker.__exit__()
 
     def get(
-        self, logger: logging.Logger, key: str, output_types: Env.Bindings[Type.Base]
+        self,
+        key: str,
+        output_types: Env.Bindings[Type.Base],
+        logger: Optional[logging.Logger] = None,
     ) -> Optional[Env.Bindings[Value.Base]]:
         """
         Resolve cache key to call outputs, if available, or None. When matching outputs are found,
@@ -57,7 +42,9 @@ class CallCache:
         """
         raise NotImplementedError()
 
-    def put(self, logger: logging.Logger, key: str, outputs: Env.Bindings[Value.Base]) -> None:
+    def put(
+        self, key: str, outputs: Env.Bindings[Value.Base], logger: Optional[logging.Logger] = None
+    ) -> None:
         """
         Store call outputs for future reuse
         """
@@ -105,17 +92,18 @@ class CallCache:
                     )
         return None
 
-    def get_download(self, logger: logging.Logger, uri: str) -> Optional[str]:
+    def get_download(self, uri: str, logger: Optional[logging.Logger] = None) -> Optional[str]:
         """
         Return filename of the cached download of uri, if available. If so then opens a shared
         flock on the local file, which will remain for the life of the CallCache object.
         """
+        logger = logger.getChild("CallCache") if logger else self._logger
         p = self.download_path(uri)
         if not (self._cfg["download_cache"].get_bool("get") and p and os.path.isfile(p)):
             logger.debug(_("no download cache hit", uri=uri, cache_path=p))
             return None
         try:
-            self._flock([p])
+            self._flocker.flock(p, update_atime=True)
             logger.info(_("found in download cache", uri=uri, cache_path=p))
             return p
         except Exception as exn:
@@ -129,11 +117,14 @@ class CallCache:
             )
             return None
 
-    def put_download(self, logger: logging.Logger, uri: str, filename: str) -> str:
+    def put_download(
+        self, uri: str, filename: str, logger: Optional[logging.Logger] = None,
+    ) -> str:
         """
         Move the downloaded file to the cache location & return the new path; or if the uri isn't
         cacheable, return the given path.
         """
+        logger = logger.getChild("CallCache") if logger else self._logger
         ans = filename
         if self._cfg["download_cache"].get_bool("put"):
             p = self.download_path(uri)
@@ -142,29 +133,5 @@ class CallCache:
                 os.rename(filename, p)
                 logger.info(_("stored in download cache", uri=uri, cache_path=p))
                 ans = p
-        self._flock([ans])
+        self._flocker.flock(ans, update_atime=True)
         return ans
-
-
-@contextmanager
-def _open_and_flock(
-    filename: str, mode: str = "rb", exclusive: bool = False, wait: bool = False
-) -> Iterator[IO[Any]]:
-    """
-    context manager yields an open BinaryIO/TextIO with a flock on the file, also updating atime
-    """
-    while True:
-        with open(filename, mode) as openfile:
-            op = fcntl.LOCK_EX if exclusive else fcntl.LOCK_SH
-            if not wait:
-                op |= fcntl.LOCK_NB
-            fcntl.flock(openfile, op)
-            # verify the hardlink didn't change in between our open & flock syscalls
-            filename_st = os.stat(filename)
-            file_st = os.stat(openfile.fileno())
-            if filename_st.st_dev == file_st.st_dev and filename_st.st_ino == file_st.st_ino:
-                # touch -a
-                os.utime(openfile.fileno(), ns=(int(time.time() * 1e9), file_st.st_mtime_ns))
-                yield openfile
-                return
-        # the flock should expire automatically when we close openfile
