@@ -17,11 +17,13 @@ security credentials.
 """
 import os
 import logging
+import traceback
 import importlib_metadata
 from typing import Optional, List, Generator, Dict, Any, Tuple, Callable
 from . import config
 from .cache import CallCache
 from .._util import compose_coroutines
+from .._util import StructuredLogMessage as _
 
 # WDL tasks for downloading a file based on its URI scheme
 
@@ -99,7 +101,7 @@ def run(cfg: config.Loader, logger: logging.Logger, uri: str, **kwargs) -> str:
     useful in particular.
     """
 
-    from . import run_local_task, RunFailed, DownloadFailed, Terminated
+    from . import run_local_task, RunFailed, DownloadFailed, Terminated, error_json
     from .. import parse_tasks, values_from_json, values_to_json
 
     gen = _downloader(cfg, uri)
@@ -114,7 +116,9 @@ def run(cfg: config.Loader, logger: logging.Logger, uri: str, **kwargs) -> str:
                 task = parse_tasks(task_wdl, version="1.0")[0]  # pyre-ignore
                 task.typecheck()
                 inputs = values_from_json(inputs, task.available_inputs)  # pyre-ignore
-                subdir, outputs_env = run_local_task(cfg, task, inputs, **kwargs)
+                subdir, outputs_env = run_local_task(
+                    cfg, task, inputs, run_id=("download-" + task.name), **kwargs
+                )
 
                 recv = cor.send(
                     {"outputs": values_to_json(outputs_env), "dir": subdir}  # pyre-ignore
@@ -128,8 +132,10 @@ def run(cfg: config.Loader, logger: logging.Logger, uri: str, **kwargs) -> str:
         if isinstance(exn.__cause__, Terminated):
             raise exn.__cause__ from None
         raise DownloadFailed(uri) from exn.__cause__
-    except:
-        raise DownloadFailed(uri)
+    except Exception as exn:
+        logger.debug(traceback.format_exc())
+        logger.error(_("downloader error", uri=uri, **error_json(exn)))
+        raise DownloadFailed(uri) from exn
 
 
 def run_cached(
@@ -139,7 +145,7 @@ def run_cached(
     Cached download logic: returns the file from the cache if available; otherwise, runs the
     download and puts it into the cache before returning
     """
-    cached = cache.get_download(logger, uri)
+    cached = cache.get_download(uri, logger=logger)
     if cached:
         return True, cached
     if not cfg["download_cache"].get_bool("put") or not cache.download_path(uri):
@@ -147,7 +153,7 @@ def run_cached(
     # run the download within the cache directory
     run_dir = os.path.join(cfg["download_cache"]["dir"], "ops")
     filename = run(cfg, logger, uri, run_dir=run_dir, **kwargs)
-    return False, cache.put_download(logger, uri, os.path.realpath(filename))
+    return False, cache.put_download(uri, os.path.realpath(filename), logger=logger)
 
 
 def gsutil_downloader(
@@ -158,6 +164,9 @@ def gsutil_downloader(
 
     TODO: adopt security credentials from runtime environment
     """
+    if uri == "gs://8675309":
+        # hook for test coverage of exception handler
+        raise RuntimeError("don't change your number")
     wdl = r"""
     task gsutil_cp {
         input {
