@@ -16,6 +16,7 @@ import threading
 import shutil
 import shlex
 import re
+import socket
 from abc import ABC, abstractmethod
 from typing import Tuple, List, Dict, Optional, Callable, Iterable, Set, Any
 import docker
@@ -261,7 +262,7 @@ class TaskContainer(ABC):
         return None
 
 
-class LocalSwarmContainer(TaskContainer):
+class SwarmContainer(TaskContainer):
     """
     TaskContainer docker (swarm) runtime
     """
@@ -272,7 +273,7 @@ class LocalSwarmContainer(TaskContainer):
     @classmethod
     def global_init(cls, cfg: config.Loader, logger: logging.Logger) -> None:
         client = docker.from_env()
-        detector = None
+        worker_nodes = []
         try:
             logger.debug("dockerd :: " + json.dumps(client.version())[1:-1])
 
@@ -281,64 +282,95 @@ class LocalSwarmContainer(TaskContainer):
             while True:
                 info = client.info()
                 if "Swarm" in info and "LocalNodeState" in info["Swarm"]:
+                    logger.debug(_("swarm info", **info["Swarm"]))
                     state = info["Swarm"]["LocalNodeState"]
 
                 # https://github.com/moby/moby/blob/e7b5f7dbe98c559b20c0c8c20c0b31a6b197d717/api/types/swarm/swarm.go#L185
-                if state == "inactive":
+                if state == "active":
+                    if info["Swarm"]["ControlAvailable"]:
+                        worker_nodes = [
+                            node
+                            for node in client.nodes.list()
+                            if node.attrs["Spec"]["Availability"] == "active"
+                            and node.attrs["Status"]["State"] == "ready"
+                        ]
+                        if worker_nodes:
+                            break
+                    else:
+                        logging.warning(
+                            "this host is a docker swarm worker but not a manager; WDL task scheduling requires manager access"
+                        )
+                elif state == "inactive" and cfg["docker_swarm"].get_bool("auto_init"):
                     logger.warning(
                         "docker swarm is inactive on this host; performing `docker swarm init --advertise-addr 127.0.0.1 --listen-addr 127.0.0.1`"
                     )
                     client.swarm.init(advertise_addr="127.0.0.1", listen_addr="127.0.0.1")
-                elif state == "active":
-                    break
-                else:
-                    logger.notice(  # pyre-ignore
-                        _("waiting for docker swarm to become active", state=state)
+
+                logger.notice(  # pyre-fixme
+                    _(
+                        "waiting for local docker swarm manager & worker(s)",
+                        manager=state,
+                        workers=len(worker_nodes),
                     )
-                    time.sleep(2)
+                )
+                time.sleep(2)
 
             miniwdl_services = [
                 d
                 for d in [s.attrs for s in client.services.list()]
                 if "Spec" in d and "Labels" in d["Spec"] and "miniwdl_run_id" in d["Spec"]["Labels"]
             ]
-            if miniwdl_services:
+            if miniwdl_services and cfg["docker_swarm"].get_bool("auto_init"):
                 logger.warning(
                     "docker swarm lists existing miniwdl-related services. This is normal if other miniwdl processes are running concurrently; otherwise, stale state could interfere with this run. To reset it, `docker swarm leave --force`"
                 )
+        finally:
+            client.close()
 
-            # Detect CPUs & memory available to Docker containers on the local host. These limits
-            # may be less than multiprocessing.cpu_count() and psutil.virtual_memory().total; in
-            # particular on macOS, where Docker containers run in a virtual machine with limited
-            # resources.
-            detector = client.containers.run(
-                "alpine:3",
-                name=f"wdl-detector-{os.getpid()}",
-                command=["/bin/ash", "-c", "nproc && free -b | awk '/^Mem:/{print $2}'"],
-                log_config=docker.types.LogConfig(type=docker.types.LogConfig.types.JSON),
-                detach=True,
-            )
-            status = detector.wait()
-            assert (
-                isinstance(status, dict) and status.get("StatusCode", -1) == 0
-            ), f"failed to detect host docker limits: f{str(status)}"
-            stdout = detector.logs(stdout=True)
-            logger.debug(_("docker resource detection", stdout=str(stdout)))
-            stdout = stdout.decode("utf-8").strip().split("\n")
-            assert len(stdout) == 2
-            ans = (int(stdout[0]), int(stdout[1]))
-            logger.info(
+        # Detect swarm's CPU & memory resources. Even on a localhost swarm, these may be less than
+        # multiprocessing.cpu_count() and psutil.virtual_memory().total; in particular on macOS,
+        # where Docker containers run in a virtual machine with limited resources.
+        resources_max_mem = {}
+        total_NanoCPUs = 0
+        total_MemoryBytes = 0
+
+        for node in worker_nodes:
+            logger.debug(
                 _(
-                    "detected host resources available for Docker containers",
-                    cpu=ans[0],
-                    mem_bytes=ans[1],
+                    "swarm worker",
+                    ID=node.attrs["ID"],
+                    Spec=node.attrs["Spec"],
+                    Hostname=node.attrs["Description"]["Hostname"],
+                    Resources=node.attrs["Description"]["Resources"],
+                    Status=node.attrs["Status"],
                 )
             )
-            cls._limits = {"cpu": ans[0], "mem_bytes": ans[1]}
-        finally:
-            if detector:
-                detector.remove()
-            client.close()
+            resources = node.attrs["Description"]["Resources"]
+            total_NanoCPUs += resources["NanoCPUs"]
+            total_MemoryBytes += resources["MemoryBytes"]
+            if (
+                not resources_max_mem
+                or resources["MemoryBytes"] > resources_max_mem["MemoryBytes"]
+                or (
+                    resources["MemoryBytes"] == resources_max_mem["MemoryBytes"]
+                    and resources["NanoCPUs"] > resources_max_mem["NanoCPUs"]
+                )
+            ):
+                resources_max_mem = resources
+
+        max_cpu = int(resources_max_mem["NanoCPUs"] / 1_000_000_000)
+        max_mem = resources_max_mem["MemoryBytes"]
+        logger.notice(  # pyre-ignore
+            _(
+                "docker swarm resources",
+                workers=len(worker_nodes),
+                max_cpus=max_cpu,
+                max_mem_bytes=max_mem,
+                total_cpus=int(total_NanoCPUs / 1_000_000_000),
+                total_mem_bytes=total_MemoryBytes,
+            )
+        )
+        cls._limits = {"cpu": max_cpu, "mem_bytes": max_mem}
 
     @classmethod
     def detect_resource_limits(cls, cfg: config.Loader, logger: logging.Logger) -> Dict[str, int]:
@@ -392,14 +424,13 @@ class LocalSwarmContainer(TaskContainer):
         client = docker.from_env(timeout=900)
         resources, user, groups = self.misc_config(logger, client, cpu, memory)
         svc = None
+        exit_code = None
         try:
             # run container as a transient docker swarm service, letting docker handle the resource
             # scheduling (waiting until requested # of CPUs are available).
             kwargs = {
                 # unique name with some human readability; docker limits to 63 chars (issue #327)
-                "name": f"wdl-{os.getpid()}-{LocalSwarmContainer._id_counter.next()}-{self.run_id}"[
-                    :63
-                ],
+                "name": f"wdl-{os.getpid()}-{SwarmContainer._id_counter.next()}-{self.run_id}"[:63],
                 "command": [
                     "/bin/bash",
                     "-c",
@@ -420,7 +451,6 @@ class LocalSwarmContainer(TaskContainer):
             svc = client.services.create(self.image_tag, **kwargs)
             logger.debug(_("docker service", name=svc.name, id=svc.short_id))
 
-            exit_code = None
             # stream stderr into log
             with PygtailLogger(logger, os.path.join(self.host_dir, "stderr.txt")) as poll_stderr:
                 # poll for container exit
@@ -430,9 +460,7 @@ class LocalSwarmContainer(TaskContainer):
                     if terminating():
                         raise Terminated() from None
                     if "running" in self._observed_states:
-                        if not running:
-                            logger.notice("container running")  # pyre-fixme
-                            running = True
+                        running = True
                         poll_stderr()
                     exit_code = self.poll_service(logger, svc)
                 logger.debug(
@@ -442,7 +470,6 @@ class LocalSwarmContainer(TaskContainer):
                         stderr=list(msg.decode().rstrip() for msg in svc.logs(stderr=True)),
                     )
                 )
-                logger.info(_("docker exit", code=exit_code))
 
             # retrieve and check container exit status
             assert isinstance(exit_code, int)
@@ -453,7 +480,7 @@ class LocalSwarmContainer(TaskContainer):
                     svc.remove()
                 except:
                     logger.exception("failed to remove docker service")
-                self.chown(logger, client)
+                self.chown(logger, client, exit_code == 0)
             try:
                 client.close()
             except:
@@ -547,7 +574,7 @@ class LocalSwarmContainer(TaskContainer):
         if tasks:
             assert len(tasks) == 1, "docker service should have at most 1 task"
             status = tasks[0]["Status"]
-            logger.debug(_("docker task", id=tasks[0]["ID"], status=status))
+            logger.debug(_("docker task status", **status))
         else:
             assert (
                 len(self._observed_states or []) <= 1
@@ -556,7 +583,13 @@ class LocalSwarmContainer(TaskContainer):
         # log each new state
         assert isinstance(self._observed_states, set)
         if status["State"] not in self._observed_states:
-            logger.info(_("docker task transition", state=status["State"]))
+            loginfo = {"service": svc.short_id}
+            if tasks:
+                loginfo["task"] = tasks[0]["ID"][:10]
+                if "NodeID" in tasks[0]:
+                    loginfo["node"] = tasks[0]["NodeID"][:10]
+            method = logger.notice if status["State"] == "running" else logger.info  # pyre-fixme
+            method(_(f"docker task {status['State']}", **loginfo))
             self._observed_states.add(status["State"])
 
         # https://docs.docker.com/engine/swarm/how-swarm-mode-works/swarm-task-states/
@@ -565,7 +598,9 @@ class LocalSwarmContainer(TaskContainer):
             exit_code = status["ContainerStatus"]["ExitCode"]
             assert isinstance(exit_code, int)
             if exit_code != 0 or status["State"] == "complete":
-                logger.info(_("docker task exit", state=status["State"], exit_code=exit_code))
+                logger.notice(  # pyre-fixme
+                    _("docker task exit", state=status["State"], exit_code=exit_code)
+                )
                 return exit_code
 
         if status["State"] in ["failed", "rejected", "orphaned", "remove"]:
@@ -576,7 +611,7 @@ class LocalSwarmContainer(TaskContainer):
 
         return None
 
-    def chown(self, logger: logging.Logger, client: docker.DockerClient) -> None:
+    def chown(self, logger: logging.Logger, client: docker.DockerClient, success: bool) -> None:
         """
         After task completion, chown all files in the working directory to the invoking user:group,
         instead of leaving them frequently owned by root or some other arbitrary user id (image-
@@ -584,7 +619,6 @@ class LocalSwarmContainer(TaskContainer):
         alternatives and their problems.
         """
         if not self.cfg["task_runtime"].get_bool("as_user") and (os.geteuid() or os.getegid()):
-            t_0 = time.monotonic()
             script = f"""
             chown -RP {os.geteuid()}:{os.getegid()} {shlex.quote(os.path.join(self.container_dir, 'work'))}
             """.strip()
@@ -595,7 +629,7 @@ class LocalSwarmContainer(TaskContainer):
                 try:
                     chowner = client.containers.run(
                         "alpine:3",
-                        name=f"wdl-chown-{os.getpid()}-{LocalSwarmContainer._id_counter.next()}-{self.run_id}"[
+                        name=f"wdl-chown-{os.getpid()}-{SwarmContainer._id_counter.next()}-{self.run_id}"[
                             :63
                         ],
                         command=["/bin/ash", "-c", script],
@@ -611,16 +645,9 @@ class LocalSwarmContainer(TaskContainer):
                     if chowner:
                         chowner.remove()
             except:
-                logger.exception("post-task chown failed")
-            finally:
-                t_delta = time.monotonic() - t_0
-                if t_delta >= 60:
-                    logger.warning(
-                        _(
-                            "post-task chown was slow (may indicate excessive file count and/or IOPS exhaustion)",
-                            seconds=int(t_delta),
-                        )
-                    )
+                if success:
+                    raise
+                logger.exception("post-task chown also failed")
 
 
 def run_local_task(
@@ -630,6 +657,7 @@ def run_local_task(
     run_id: Optional[str] = None,
     run_dir: Optional[str] = None,
     logger_prefix: Optional[List[str]] = None,
+    _run_id_stack: Optional[List[str]] = None,
     _cache: Optional[CallCache] = None,
     _plugins: Optional[List[Callable[..., Any]]] = None,
 ) -> Tuple[str, Env.Bindings[Value.Base]]:
@@ -647,6 +675,7 @@ def run_local_task(
 
     # provision run directory and log file
     run_id = run_id or task.name
+    run_id_stack = (_run_id_stack or []) + [run_id]
     run_dir = provision_run_dir(task.name, run_dir)
     write_values_json(inputs, os.path.join(run_dir, "inputs.json"))
 
@@ -665,15 +694,20 @@ def run_local_task(
             )
         )
         logger.info(_("thread", ident=threading.get_ident()))
-        cache = _cache or CallCache(cfg)
 
         try:
+            cache = _cache or CallCache(cfg, logger)
+
             # start plugin coroutines and process inputs through them
             with compose_coroutines(
                 [
-                    (lambda kwargs: cor(cfg, logger, run_id, run_dir, task, **kwargs))
+                    (
+                        lambda kwargs, cor=cor: cor(
+                            cfg, logger, run_id_stack, run_dir, task, **kwargs
+                        )
+                    )
                     for cor in (
-                        [cor for _, cor in sorted(config.load_plugins(cfg, "task"))]
+                        [cor2 for _, cor2 in sorted(config.load_plugins(cfg, "task"))]
                         + (_plugins or [])
                     )
                 ],
@@ -688,7 +722,7 @@ def run_local_task(
                 )
 
                 # create appropriate TaskContainer
-                container = LocalSwarmContainer(cfg, run_id, run_dir)
+                container = SwarmContainer(cfg, run_id, run_dir)
 
                 # evaluate input/postinput declarations, including mapping from host to
                 # in-container file paths
@@ -716,12 +750,19 @@ def run_local_task(
                 # evaluate output declarations
                 outputs = _eval_task_outputs(logger, task, container_env, container)
 
-                # make sure everything will be accessible to downstream tasks
-                chmod_R_plus(container.host_dir, file_bits=0o660, dir_bits=0o770)
+                # create output_links
+                outputs = link_outputs(
+                    outputs, run_dir, hardlinks=cfg["file_io"].get_bool("output_hardlinks")
+                )
 
-                # process outputs through plugins & create output_links
+                # process outputs through plugins
                 recv = plugins.send({"outputs": outputs})
-                outputs = link_outputs(recv["outputs"], run_dir)
+                outputs = recv["outputs"]
+
+                # clean up, if so configured, and make sure output files will be accessible to
+                # downstream tasks
+                _delete_work(cfg, logger, run_dir, True)
+                chmod_R_plus(run_dir, file_bits=0o660, dir_bits=0o770)
 
                 # write outputs.json
                 write_values_json(
@@ -741,6 +782,10 @@ def run_local_task(
             except Exception as exn2:
                 logger.debug(traceback.format_exc())
                 logger.critical(_("failed to write error.json", dir=run_dir, message=str(exn2)))
+            try:
+                _delete_work(cfg, logger, run_dir, False)
+            except:
+                logger.exception("delete_work also failed")
             raise wrapper from exn
 
 
@@ -908,7 +953,7 @@ def _eval_task_runtime(
     if "docker" in runtime_values:
         ans["docker"] = runtime_values["docker"].coerce(Type.String()).value
 
-    host_limits = LocalSwarmContainer.detect_resource_limits(cfg, logger)
+    host_limits = SwarmContainer.detect_resource_limits(cfg, logger)
     if "cpu" in runtime_values:
         cpu_value = runtime_values["cpu"].coerce(Type.Int()).value
         assert isinstance(cpu_value, int)
@@ -974,7 +1019,7 @@ def _try_task(
 
     while True:
         # copy input files, if needed
-        if cfg["task_io"].get_bool("copy_input_files"):
+        if cfg["file_io"].get_bool("copy_input_files"):
             container.copy_input_files(logger)
 
         try:
@@ -1064,7 +1109,9 @@ def _eval_task_outputs(
     return outputs
 
 
-def link_outputs(outputs: Env.Bindings[Value.Base], run_dir: str) -> Env.Bindings[Value.Base]:
+def link_outputs(
+    outputs: Env.Bindings[Value.Base], run_dir: str, hardlinks: bool = False
+) -> Env.Bindings[Value.Base]:
     """
     Following a successful run, the output files may be scattered throughout a complex directory
     tree used for execution. To help navigating this, generate a subdirectory of the run directory
@@ -1077,10 +1124,10 @@ def link_outputs(outputs: Env.Bindings[Value.Base], run_dir: str) -> Env.Binding
             if os.path.isfile(v.value):
                 hardlink = os.path.realpath(v.value)
                 assert os.path.isfile(hardlink)
-                symlink = os.path.join(dn, os.path.basename(v.value))
+                newlink = os.path.join(dn, os.path.basename(v.value))
                 os.makedirs(dn, exist_ok=False)
-                os.symlink(hardlink, symlink)
-                v.value = symlink
+                (os.link if hardlinks else os.symlink)(hardlink, newlink)
+                v.value = newlink
         # recurse into compound values
         elif isinstance(v, Value.Array) and v.value:
             d = int(math.ceil(math.log10(len(v.value))))  # how many digits needed
@@ -1115,16 +1162,32 @@ def link_outputs(outputs: Env.Bindings[Value.Base], run_dir: str) -> Env.Binding
                 v.value[key] = map_files(v.value[key], os.path.join(dn, key))
         return v
 
+    os.makedirs(os.path.join(run_dir, "output_links"), exist_ok=False)
     return outputs.map(
         lambda binding: Env.Binding(
             binding.name,
             map_files(
-                copy.deepcopy(binding.value), os.path.join(run_dir, "output_links", binding.name)
+                copy.deepcopy(binding.value), os.path.join(run_dir, "output_links", binding.name),
             ),
         )
     )
 
 
+def _delete_work(cfg: config.Loader, logger: logging.Logger, run_dir: str, success: bool) -> None:
+    opt = cfg["task_runtime"]["delete_work"].strip().lower()
+    if opt == "always" or (success and opt == "success") or (not success and opt == "failure"):
+        if success and not cfg["file_io"].get_bool("output_hardlinks"):
+            logger.warning(
+                "ignoring configuration [task_runtime] delete_work because it requires [file_io] output_hardlinks = true"
+            )
+            return
+        for dn in ["write_", "work"]:
+            dn = os.path.join(run_dir, dn)
+            if os.path.isdir(dn):
+                shutil.rmtree(dn)
+                logger.info(_("deleted working directory", dir=dn))
+
+
 class _StdLib(StdLib.Base):
     logger: logging.Logger
     container: TaskContainer
