@@ -17,7 +17,7 @@ class TestTaskRunner(unittest.TestCase):
         logging.basicConfig(level=logging.DEBUG, format='%(name)s %(levelname)s %(message)s')
         logger = logging.getLogger(cls.__name__)
         cfg = WDL.runtime.config.Loader(logger, [])
-        WDL.runtime.task.LocalSwarmContainer.global_init(cfg, logger)
+        WDL.runtime.task.SwarmContainer.global_init(cfg, logger)
 
     def setUp(self):
         self._dir = tempfile.mkdtemp(prefix="miniwdl_test_taskrun_")
@@ -32,6 +32,7 @@ class TestTaskRunner(unittest.TestCase):
             if isinstance(inputs, dict):
                 inputs = WDL.values_from_json(inputs, doc.tasks[0].available_inputs, doc.tasks[0].required_inputs)
             rundir, outputs = WDL.runtime.run_local_task(cfg, doc.tasks[0], (inputs or WDL.Env.Bindings()), run_dir=self._dir, **kwargs)
+            self._rundir = rundir
         except WDL.runtime.RunFailed as exn:
             if expected_exception:
                 self.assertIsInstance(exn.__context__, expected_exception)
@@ -645,7 +646,7 @@ class TestTaskRunner(unittest.TestCase):
         outputs = self._test_task(txt, inp)
         chk(outputs["outfiles"])
         cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()), [])
-        cfg.override({"task_io": {"copy_input_files": True}})
+        cfg.override({"file_io": {"copy_input_files": True}})
         outputs = self._test_task(txt, inp, cfg=cfg)
         chk(outputs["outfiles"])
 
@@ -800,7 +801,7 @@ class TestTaskRunner(unittest.TestCase):
                         expected_exception=WDL.runtime.task.CommandFailed)
 
         cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()), [])
-        cfg.override({"task_io": {"copy_input_files": True}})
+        cfg.override({"file_io": {"copy_input_files": True}})
         outputs = self._test_task(txt, {"files": [os.path.join(self._dir, "alyssa.txt"), os.path.join(self._dir, "ben.txt")]}, cfg=cfg)
         self.assertTrue(outputs["outfile"].endswith("alyssa2.txt"))
 
@@ -855,7 +856,8 @@ class TestTaskRunner(unittest.TestCase):
 
     @log_capture()
     def test_download_cache(self, capture):
-        cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()))
+        logger = logging.getLogger(self.id())
+        cfg = WDL.runtime.config.Loader(logger)
         cfg.override({
             "download_cache": {
                 "put": True,
@@ -884,6 +886,15 @@ class TestTaskRunner(unittest.TestCase):
         self.assertTrue("downloaded: 1" in logs[0])
         self.assertTrue("cached: 1" in logs[1])
 
+        # quick test CallCache reentrancy
+        cache = WDL.runtime.cache.CallCache(cfg, logger)
+        self.assertIsNotNone(cache.get_download("https://google.com/robots.txt", logger=logger))
+        self.assertIsNotNone(cache.get_download("https://google.com/robots.txt", logger=logger))
+        self.assertEqual(len(cache._flocker._flocks), 1)
+        with self.assertRaises(OSError):
+            cache._flocker.flock(cache.download_path("https://google.com/robots.txt"), exclusive=True)
+        del cache
+
     def test_workdir_ownership(self):
         # verify that everything within working directory is owned by the invoking user
         txt = R"""
@@ -910,7 +921,43 @@ class TestTaskRunner(unittest.TestCase):
         outputs = self._test_task(txt, {"files": [os.path.join(self._dir, "alyssa.txt"), os.path.join(self._dir, "ben.txt")]})
         self.assertEqual(len(outputs["uids"]), 1)
         self.assertEqual(outputs["uids"][0], os.geteuid())
-    
+
+    def test_delete_work(self):
+        txt = R"""
+        version 1.0
+        task xxx {
+            input {
+                Array[File] files
+            }
+            File written = write_lines(files)
+            command <<<
+                set -euxo pipefail
+                cp "~{written}" foo.txt
+                cp "~{files[0]}" bar.txt
+            >>>
+            output {
+                Array[File] outfiles = [write_lines(files), "foo.txt", "bar.txt"]
+            }
+        }
+        """
+        with open(os.path.join(self._dir, "alyssa.txt"), "w") as outfile:
+            outfile.write("Alyssa\n")
+        with open(os.path.join(self._dir, "ben.txt"), "w") as outfile:
+            outfile.write("Ben\n")
+        inputs = {"files": [os.path.join(self._dir, "alyssa.txt"), os.path.join(self._dir, "ben.txt")]}
+        cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()), [])
+        cfg.override({"file_io": {"output_hardlinks": True}, "task_runtime": {"delete_work": "success"}})
+        output = self._test_task(txt, inputs, cfg=cfg)
+        self.assertFalse(os.path.isdir(os.path.join(self._rundir, "work")))
+        self.assertFalse(os.path.isdir(os.path.join(self._rundir, "write_")))
+        for fn in output["outfiles"]:
+            self.assertTrue(os.path.isfile(fn) and not os.path.islink(fn))
+
+        cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()), [])
+        cfg.override({"file_io": {"output_hardlinks": False}, "task_runtime": {"delete_work": "success"}})
+        output = self._test_task(txt, inputs, cfg=cfg)
+        self.assertTrue(os.path.isfile(os.path.join(self._rundir, "work", "foo.txt")))
+
     def test_plugins(self):
         def my_plugin(cfg, logger, task, run_id, run_dir, **recv):
             logger = logger.getChild("my_plugin")
@@ -986,36 +1033,36 @@ class TestConfigLoader(unittest.TestCase):
         logging.basicConfig(level=logging.DEBUG, format='%(name)s %(levelname)s %(message)s')
         logger = logging.getLogger(cls.__name__)
         cfg = WDL.runtime.config.Loader(logger, [])
-        WDL.runtime.task.LocalSwarmContainer.global_init(cfg, logger)
+        WDL.runtime.task.SwarmContainer.global_init(cfg, logger)
 
     # trigger various codepaths of the config loader that wouldn't be exercised otherwise
     def test_basic(self):
         cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()), [])
-        self.assertEqual(cfg["task_io"]["copy_input_files"], "false")
-        self.assertEqual(cfg["task_io"].get_bool("copy_input_files"), False)
+        self.assertEqual(cfg["file_io"]["copy_input_files"], "false")
+        self.assertEqual(cfg["file_io"].get_bool("copy_input_files"), False)
 
         self.assertEqual(cfg["scheduler"].get_int("call_concurrency"), 0)
 
-        cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()), overrides = {"task_io":{"copy_input_files": "true"}})
-        self.assertEqual(cfg["task_io"].get_bool("copy_input_files"), True)
+        cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()), overrides = {"file_io":{"copy_input_files": "true"}})
+        self.assertEqual(cfg["file_io"].get_bool("copy_input_files"), True)
 
         with self.assertRaises(configparser.NoSectionError):
             cfg.get("bogus", "key")
         with self.assertRaises(configparser.NoOptionError):
-            cfg.get("task_io", "bogus")
-        self.assertTrue(cfg.has_option("task_io", "copy_input_files"))
+            cfg.get("file_io", "bogus")
+        self.assertTrue(cfg.has_option("file_io", "copy_input_files"))
         self.assertFalse(cfg.has_option("bogus", "key"))
-        self.assertFalse(cfg.has_option("task_io", "bogus"))
+        self.assertFalse(cfg.has_option("file_io", "bogus"))
 
         with self.assertRaises(ValueError):
-            cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()), overrides = {"task_io":{"copy_input_files": "bogus123"}})
-            cfg.get_bool("task_io", "copy_input_files")
+            cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()), overrides = {"file_io":{"copy_input_files": "bogus123"}})
+            cfg.get_bool("file_io", "copy_input_files")
 
     def test_env(self):
         with tempfile.NamedTemporaryFile(mode="w+", delete=False) as tmp:
             assert(os.path.isabs(tmp.name))
             print("""
-            [task_io]
+            [file_io]
             copy_input_files = true
             expansion = $HOME
             made_up = 42
@@ -1027,7 +1074,7 @@ class TestConfigLoader(unittest.TestCase):
             cfg = WDL.runtime.config.Loader(logging.getLogger(self.id()))
             cfg.override({"bogus": {"option2": "42"}})
             self.assertEqual(cfg["scheduler"].get_int("call_concurrency"), 4)
-            self.assertEqual(cfg["task_io"].get_bool("copy_input_files"), True)
+            self.assertEqual(cfg["file_io"].get_bool("copy_input_files"), True)
             cfg.log_all()
             cfg.log_unused_options()
-            self.assertTrue(os.path.isabs(cfg["task_io"]["expansion"]))
+            self.assertTrue(os.path.isabs(cfg["file_io"]["expansion"]))
