@@ -18,7 +18,7 @@ class RunnerTestCase(unittest.TestCase):
         logging.basicConfig(level=logging.DEBUG, format='%(name)s %(levelname)s %(message)s')
         logger = logging.getLogger(cls.__name__)
         cfg = WDL.runtime.config.Loader(logger, [])
-        WDL.runtime.task.LocalSwarmContainer.global_init(cfg, logger)
+        WDL.runtime.task.SwarmContainer.global_init(cfg, logger)
 
     def setUp(self):
         """
@@ -80,6 +80,7 @@ class TestDownload(RunnerTestCase):
             "gs://gcp-public-data-landsat/LC08/01/044/034/LC08_L1GT_044034_20130330_20170310_01_T2/LC08_L1GT_044034_20130330_20170310_01_T2_MTL.txt",
             "https://raw.githubusercontent.com/chanzuckerberg/miniwdl/master/nonexistent12345.txt", "https://raw.githubusercontent.com/chanzuckerberg/miniwdl/master/tests/alyssa_ben.txt"
         ]}, expected_exception=WDL.runtime.DownloadFailed)
+        self._run(self.count_wdl, {"files": ["gs://8675309"]}, expected_exception=WDL.runtime.DownloadFailed)
 
     @log_capture()
     def test_download_cache1(self, capture):
